## A port of the Footnotes plugin from Flatpress
#
# Original author: NoWhereMan

import re

## The text that precedes the footnote list
FOOTNOTES_START = '[footnotes]'

## FlatPress plugin allows to use *strong* and _italic_ markup.
# I do not like this behavior, so I added this flat to disable it.
FOOTNOTES_MARKUP = False

## The regex used to tell where is the beginning of the footnote list
_footnotesRe = re.compile('\n'+re.escape(FOOTNOTES_START)+'\\s*(.*)',
        re.S | re.M)

## The class used to handle footnotes
#
# This class is not exactly a port of the original from Flatpress, but the code
# basically do the same things.
class Footnotes:

    ## The HTML code for references
    REF = '<sup><a id="{0}" href="#{1}" title="Note {2}">{2}</a></sup>'

    ## The HTML code for notes
    NOTE = '<li>{0} \n'+\
            '<a id="{1}-fn-{2}" href="#{1}-rel-{2}" title="Back {2}">^top</a>'+\
            '</li>\n'

    ## Saves the id to the instance of the class
    #
    # Saving the id in this way allows having easier code in the replacement of
    # Wordpress hook, because it allows having only the text as parameter in
    # apply.
    #
    # \param id_ An id that is used only to have different footnotes in one page
    #            otherwise different HTML elements would have the same id
    def __init__(self, id_):
        self.id = id_

    ## Handle the footnotes in an entry
    #
    # \param The text of the entry
    # \return The text with parsed footnotes
    def apply(self, text):
        if FOOTNOTES_MARKUP:
            # *STRONG* emphasis
            text = re.sub('(?<!\S)\*(?=\S) (?! \*) (.+?) (?<=\S) \*(?!>\w)',
                    r'<strong>\1</strong>', text, 0, re.X | re.S)
            # _emphasis_ (italic)
            text = re.sub('(?<!\S)\_(?=\S) (?! \_) (.+?) (?<=\S) \_(?!>\w)',
                    r'<em>\1</em>', text, 0, re.X | re.S)

        text = _footnotesRe.sub(self.footnotes, text)
        text = re.sub(r'\[([0-9]+)\]', self.references, text)

        return text

    ## The callback to pass to re.sub to create the list of footnotes
    #
    # \param m The matches of re.sub
    # \return The content to replace the footnote list with
    def footnotes(self, m):
        ret = '<div class="footnotes">\n<h4>Footnotes</h4>\n<ol>\n';

        lines = re.split(r'\[([0-9]+)\]', m.group(1));
        lines.pop(0)

        for i in range(0, len(lines), 2):
            n = lines[i]
            ret += self.NOTE.format(lines[i+1].strip(), self.id, n)

        ret += '\t</ol>\n</div>'

        return ret

    ## The callback to pass to re.sub to create references in the text
    #
    # \param m The matches of re.sub
    # \return The content to replace a reference with
    def references(self, m):
        n = m.group(1);

        href_rel = '{}-rel-{}'.format(self.id, n)
        href_note = '{}-fn-{}'.format(self.id, n)

        return self.REF.format(href_rel, href_note, n)
