## Handle BBCode in entries, pages and comments.
#
# It depends on bbcode by Dan Watson (@dcwatson on GitHub).
# Optionally this class can use Pillow to get the size of local images and
# iptcinfo to load caption and name embedded in JPEG files.
# In particular I managed to use iptcinfo3 fork by @crccheck.
#
# I tried to keep the behavior of FlatPress, but I have also added some useful
# attributes to images (i.e. a better way to handle alt and title and the
# possibility to set images as part of the same "gallery"), and I reworked the
# video tag (YouTube support only, with iframes in "enhanched privacy" mode,
# and both traditional and youtu.be links).
#
# Finally I added the stubs to support other features and plugins, like URL
# remapping (of "attachs/" and "images/"), quick gallery, lightbox, syntax
# highlighting of code and other functionalities in general.
# However these depend on preferences of the developer (e.g. the library to use
# for syntax highlighting or for lightbox) or on the decisions for the system
# (e.g. where to store images and attachments).
# So I created a base class (BBPlugins) that one can extend to implement all
# these things.
#
# Legal stuff: most of this plugin is based on the BBCode plugin from FlatPress,
# that was somehow "linked" (whatever it means) to GPLv2 libraries, and the
# original author of the PHP version of this plugin is NoWhereMan.
# This file uses also other GPL licensed libraries.
# Therefore I have to release this file under the GNU General Public License v2.
# I (Pier Angelo Vendrame, a.k.a PieroV) am the author of this very file.
# If you manage to remove any constraint on GPL, you can do whatever you want
# with this file, with any license.

# Most important dependency
import bbcode

# To get the basename of images and use it as default alt/title attribute
import os.path

# To get the video id from YouTube URLs
from urllib.parse import urlparse, parse_qs

try:
    # To get size of local images
    from PIL import Image
except:
    Image = None

try:
    # To get alt/title attributes embedded in JPEG files
    from iptcinfo3 import IPTCInfo
except:
    IPTCInfo = None

## The class that manages the BBCode for entries and comments, in FP style.
class FpBBCode:

    ## The maximum length for URL to be printed completely
    URL_MAXLEN = 40

    ## Simple tags (tags that will be replaced with the same HTML tag) that can
    # be used both in entries and comments
    COMMON = ['strong', 'em', 'ins', 'del', 'pre']

    ## Simple tags that can be used only in entries
    ENTRY_SIMPLE = ['h2', 'h3', 'h4', 'h5', 'h6']

    ## Tags that correspond to other HTML tags and can be used both in entries
    # and comments
    COMMON_REPLACE = {'b': 'strong', 'i': 'em'}

    ## Tags that correspond to other HTML tags and can be used only in entries
    ENTRY_REPLACE = {'*': 'li'}

    ## Tags that have an opening HTML tag different from the closing one.
    # As in other cases, common means that they can be used both for entries
    # and for comments
    COMMON_VALUE_REPLACE = {
            '<blockquote><p>%(value)s</p></blockquote>': ['blockquote',
            'quote'],
            '<span style="text-decoration: line-through;">%(value)s</span>':
            ['s', 'strike'],
            '<span style="text-decoration: underline;">%(value)s</span>':
            ['u']
        }

    ## Tags that need a parameter. Available only for entries, and for this
    # reason, we rely on the author of the posts to write good HTML, i.e. we do
    # not escape it.
    ENTRY_PARAMETRIZED = {
            'color': '<span style="color: {};">{}</span>',
            'size':  '<span style="font-size: {};">{}</span>',
            'align': '<div style="text-align: {};">{}</div>',
    }

    ## Initialize the parsers
    #
    # \param plugins The class that should be used to implement additional
    #                features (see at the beginning of the file)
    def __init__(self, plugins=None):
        self.postParser = bbcode.Parser(
                newline='\n',
                install_defaults=False,
                escape_html=False,
                replace_links=False,
                replace_cosmetic=False
            )

        self.commentParser = bbcode.Parser(
                newline='\n',
                install_defaults=False,
                escape_html=True,
                replace_links=True,
                replace_cosmetic=False
            )

        post = self.postParser
        comm = self.commentParser
        self.registerSimple([post, comm], self.COMMON, self.COMMON_REPLACE)
        self.registerSimple([post], self.ENTRY_SIMPLE, self.ENTRY_REPLACE)

        self.registerParametrized()
        self.registerCallback()

        post.add_simple_formatter('hr', '<hr />', standalone=True)

        if plugins:
            self.plugins = plugins()
        else:
            self.plugins = BBPlugins()

        self.plugins.registerTags(post, comm)

    ## Register simple tags
    #
    # \param parsers The parser to register the tag to
    # \param simplest The tags that have the same name as HTML
    # \params replace The tags that have another name in HTML
    def registerSimple(self, parsers, simplest, replace):
        simple = {tag: tag for tag in simplest}
        simple.update(replace)
        for tag, html in simple.items():
            for p in parsers:
                p.add_simple_formatter(tag, '<{0}>%(value)s</{0}>'.format(html))

    ## Register tags that are parametrized
    def registerParametrized(self):
        post = self.postParser
        comm = self.commentParser

        for replacement, tags in self.COMMON_VALUE_REPLACE.items():
            for tag in tags:
                post.add_simple_formatter(tag, replacement)
                comm.add_simple_formatter(tag, replacement)

        for tag, replacement in self.ENTRY_PARAMETRIZED.items():
            self.parametrizedReplacer(post, tag, replacement)

    ## Register tags that need a callback
    def registerCallback(self):
        tags = {
            'list': self.list,
            'url': self.link,
        }

        standalone = {
            'img': self.image,
            'video': self.video,
            'gallery': self.gallery,
        }

        for tag, callback in tags.items():
            self.postParser.add_formatter(tag, callback)

        for tag, callback in standalone.items():
            self.postParser.add_formatter(tag, callback, standalone=True)

        # Code is different because it needs custom options
        for parser in (self.postParser, self.commentParser):
            # Disable BBCode parsing inside, but allow HTML, so that one can
            # have a syntax highlighter, e.g. Pygments
            parser.add_formatter('code', self.code, render_embedded=False,
                    escape_html=False)

    ## Return a callback to use for a parametrized replacer
    #
    # \param parser The parser to add this tag to
    # \param tagName The name of the tag to parse
    # \param replacement The replacement for the tag
    def parametrizedReplacer(self, parser, tagName, replacement):
        def _callback(tagName, value, options, parent, context):
            if tagName not in options:
                return value
            return replacement.format(options[tagName], value)
        parser.add_formatter(tagName, _callback)

    ## Handle a list
    def list(self, tagName, value, options, parent, context):
        tag = 'ol' if tagName in options and options[tagName] == '#' else 'ul'
        return '<{0}>{1}</{0}>'.format(tag, value)

    ## Handle an image
    def image(self, tagName, value, options, parent, context):
        if tagName not in options:
            return '[missing image name]'

        fname = options[tagName]
        src = self.plugins.remapUrl(fname)
        basename = os.path.basename(fname)
        info = self.plugins.imageInfo(fname)

        # Problem (maybe): this does not allow blank alt/title
        titles = [
                options.get('title'),
                info.get('title'),
                options.get('alt'),
                info.get('alt'),
                basename
            ]
        title = None
        while not title and titles:
            title = titles.pop(0)

        alts = [
                options.get('alt'),
                info.get('alt'),
                title
            ]
        alt = None
        while not alt and alts:
            alt = alts.pop(0)

        # Better safe than sorry
        alt = _entities(alt) if alt else ''
        title = _entities(title) if title else ''

        width = ''
        height = ''
        scale = options.get('scale')

        # Note that popup=False will be handled as popup=True, but that is
        # exactly what happens in the original file...
        popup = bool(options.get('popup'))

        if 'width' in options and 'height' in options:
            # We use style instead of classic width and height, therefore we
            # have to esplicitly specify pixels... However this allows also
            # to use a relative unit (e.g. em) and auto for the other dimension
            if options['width'].isdigit():
                options['width'] += 'px'
            if options['height'].isdigit():
                options['height'] += 'px'

            width  = options['width']
            height = options['height']
        else:
            for x in ('width', 'height'):
                if x in options and x in info and info[x]:
                    # Check both that width/height exists and it is not 0
                    try:
                        val = float(options[x])
                        scale = val / info[x]
                    except:
                        pass

        # Scale on FP has more priority than explicit width and height!
        if scale and 'width' in info and 'height' in info:
            if isinstance(scale, str):
                if scale[-1] == '%':
                    scale = scale[0:-1]
                try:
                    scale = float(scale) / 100
                except:
                    scale = None
            # We assume that it is already a float otherwise

            width = str(round(scale * info['width'])) + 'px'
            height = str(round(scale * info['height'])) + 'px'

            # Scale implies always a popup!
            popup = True

        if options.get('float') in ('left', 'right'):
            float_ = 'float'+options['float']
        else:
            float_ = 'center'

        if popup:
            # This was a modification I had made
            popupTriple = self.plugins.lightbox(src, options.get('gallery'))
        else:
            popupTriple = ('', '', '')

        if scale:
            thumbUrl = self.plugins.thumbnail(src, scale)

        if width or height:
            style = ' style="'
            if width:
                style += _entities('width: {};'.format(width))
            if height:
                style += _entities('height: {};'.format(height))
            style += '"'
        else:
            style = ''

        tpl = '{}<img src="{}" alt="{}" title="{}" class="{}"{}{} />{}'
        return tpl.format(popupTriple[0], _entities(src), alt, title, float_,
                style, popupTriple[1], popupTriple[2])

    ## Handle a link
    def link(self, tagName, value, options, parent, context):
        url = _entities(options.get(tagName, value))

        if tagName not in options:
            if len(value) > self.URL_MAXLEN:
                half = self.URL_MAXLEN // 2
                value = _entities(value[0:half])+'&hellip;'+ \
                        _entities(value[-half:])
            else:
                # Avoid encoding &hellip;
                value = _entities(value)

        url = self.plugins.remapUrl(url)

        if url.find('://') != -1:
            options['class'] = 'externlink '+options.get('class', '')

        # Allow some other attributes than FlatPress
        attrs = ''
        for tag in ('rel', 'class', 'title'):
            if tag in options:
                attrs += ' {}="{}"'.format(tag, _entities(options[tag]))

        return '<a href="{}"{}>{}</a>'.format(url, attrs, value)

    ## Handle a video
    def video(self, tagName, value, options, parent, context):
        if not tagName in options:
            return '[missing video URL]'

        url = urlparse(options[tagName])
        if url.netloc.endswith('youtube.com'):
            qs = parse_qs(url.query)
            ytId = qs.get('v', [''])[0]
        elif url.netloc.endswith('youtu.be'):
            ytId = url.path[1:]
        else:
            # RIP Google Video and flash
            return '[unsupported video]'

        if not ytId:
            return '[invalid video URL]'

        width = options.get('width', '560px')
        height = options.get('height', '315px')
        fullscreen = options.get('fullscreen', '').lower()
        if fullscreen == '0' or fullscreen == 'false' or fullscreen == 'off':
            fullscreen = ''
        else:
            fullscreen = ' allowfullscreen'

        return ('<iframe src="https://www.youtube-nocookie.com/embed/{}" ' + \
               'style="width: {}; height: {}; border-width: 0;" ' + \
               'allow="autoplay; encrypted-media"{}></iframe>').\
            format(ytId, width, height, fullscreen)

    ## Handle a gallery of images
    def gallery(self, tagName, value, options, parent, context):
        if tagName not in options:
            return '[missing gallery name]'

        name = self.plugins.sanitizeGalleryName(options[tagName])
        options['gallery'] = name
        images = self.plugins.gallery(options[tagName])

        if not images:
            return ''

        tpl = '<div class="img-gallery {}">\n{}\n</div>'
        imagesCode = ''
        for image in images:
            options[tagName] = image
            # Cannot use \n, otherwise wpautop creates a new paragraph
            imagesCode += self.image(tagName, None, options, None, None)+' '
        return tpl.format(name, imagesCode)

    ## Handle code
    def code(self, tagName, value, options, parent, context):
        lang = options.get(tagName, None)
        return self.plugins.code(value, lang)

    ## Parse an entry
    #
    # \param text The text of the entry
    # \return The rendered entry
    def entry(self, text):
        return self.postParser.format(text)

    ## Parse a comment
    #
    # \param text The text of the coment
    # \return The rendered comment
    def comment(self, text):
        return self.commentParser.format(text)

## FlatPress BBCode allows other plugins to extend it, usually through the WP
# hooks system.
#
# I have included some of these functionalities, but I do not want to implement
# the whole hook system, therefore I created this class, that can be extended
# to change the behaviour of the BBCode parser.
#
# Most of the methods of this class are just placeholders.
class BBPlugins:

    ## Remap a URL to another one, if needed
    #
    # \param url The URL to remap
    # \return The remapped URL
    def remapUrl(self, url):
        if url.startswith('www.'):
            return 'http://'+url

        tmp = url if url[0] == '/' else '/'+url
        if tmp.startswith('/attachs/') or tmp.startswith('/images/'):
            return '/fp-content' + tmp

        return url

    ## Handle a code area
    #
    # This simple implementation ignores the language, but you can use it
    # for syntax highlighting.
    #
    # \param code The text inside the code block
    # \param lang The code language, None if not specified
    # \return The modified text
    def code(self, code, lang=None):
        code = _entities(code)
        return '<pre>{}</pre>'.format(code)

    ## Get information of images
    #
    # \param filename The name of the file to get information of
    # \return A dictionary with the information we found
    def imageInfo(self, filename):
        f = self.openImage(filename)
        if not f:
            return {}

        info = {}

        if Image:
            im = Image.open(f)
            info['width']  = im.size[0]
            info['height'] = im.size[1]

        try:
            # If f was a stream, Pillow closed it, so get it again
            f = self.openImage(filename)
            iptc = IPTCInfo(f)

            if iptc['caption/abstract']:
                info['alt'] = iptc['caption/abstract'].decode()
            if iptc['object name']:
                info['title'] = iptc['object name'].decode()
        except:
            pass

        return info

    ## Return the path or a stream object to the image, or None if it has not
    # been found
    #
    # \param filename The name of the image to open
    # \return For this class always None
    def openImage(self, filename):
        return None

    ## Create the thumb of an image and return its URL
    #
    # This implementation does not to anything.
    #
    # \param src The path/URL of the original picture
    # \param scale How much the image has been rescaled
    # \return The thumb URL
    def thumbnail(self, src, scale):
        return src

    ## Return the code to add a lightbox to the image
    #
    # \param src Source URL of the image
    # \param gallery The id of the gallery (None if the image is not in a
    #                gallery)
    # \return A triple (code-before-img, attributes-in-img, code-after-img)
    def lightbox(self, src, gallery=None):
        if gallery:
            rel = ' rel="lighbox[{}]"'.format(_entities(gallery))
        else:
            rel = ' rel="lightbox"'

        return ('<a href="{}">'.format(_entities(src)), rel, '</a>')

    ## Sanitize the name of the gallery
    #
    # \param name The name of the gallery, as passed in the tag
    # \return The sanitized name
    def sanitizeGalleryName(self, name):
        # Suggestion: use something better, e.g. a slugify function
        return _entities(name)

    ## Return the list of images that compose a gallery
    #
    # \param name The gallery name
    # \return The list of images
    def gallery(self, name):
        return []

    ## Register other tags to the parsers
    #
    # \param post The post parser
    # \param comment The comment parser
    def registerTags(self, post, comment):
        # post.add_filter(...)
        # comm.add_filter(...)
        pass

## Remove HTML entities
#
# \param text The text to remove HTML entities from
# \return The encoded text
def _entities(text):
    text = text.replace('&', '&amp;')
    text = text.replace('<', '&lt;')
    text = text.replace('>', '&gt;')

    # Better replace these ones, to avoid any type beautifier
    text = text.replace('\'', '&#39;')
    text = text.replace('"', '&quot;')

    return text
