## An implementation of the readmore (manual mode) and autoreadmore plugins.
#
# Original author (for readmore): NoWhereMan

from .wpfilters import balance_tags

class Readmore:
    ## The text to use to cut the entry
    MORE_TEXT = '[more]'

    ## The template for the readmore link
    MORE_LINK = ' &hellip; <span class="readmore"><a href="{}#readmore-{}">'+\
                '[Read more]</a></span>'

    ## The template for the readmore anchor
    MORE_ANCHOR = '<a id="readmore-{}"></a>'

    ## The number of characters to keep in the preview for the automatic mode
    CHOP_AT = 1200

    ## The minimum length of the entry for the automatic mode to activate
    CHOP_MIN = 2400

    ## Initialize the "plugin"
    #
    # \param id_ An id that is used only to have different entries in one page
    #            otherwise different HTML elements would have the same id
    # \param permalink The link to the entry, used to create the readmore link
    def __init__(self, id_, permalink=None):
        self.id = id_
        self.permalink = permalink
        self.manual = False

    ## Apply the manual mode of readmore to the preview of the post, i.e. when
    # you are exploring archives.
    #
    # \param text The text of the entry
    # \return The cut text of the entry, if there was a more, otherwise the
    #         whole text
    def applyPreviewManual(self, text):
        more = text.find(self.MORE_TEXT)
        if more != -1:
            self.manual = True
            text = text[0:more]
            text += self.MORE_LINK.format(self.permalink, self.id)
            text = balance_tags(text)

        return text

    ## Replace [more] with the readmore anchor.
    #
    # \param text The text of the entry
    # \return The text with the anchor, or the original text if [more] was not
    # found
    def applyFullManual(self, text):
        if text.find(self.MORE_TEXT) != -1:
            self.manual = True
            return text.replace(self.MORE_TEXT,
                    self.MORE_ANCHOR.format(self.id), 1)
        else:
            return text

    ## Cut an entry in automatic mode
    #
    # \param text The text of the entry to cut
    # \return The cut entry
    def applyPreviewAuto(self, text):
        offset = self._autoOffset(text)
        if offset == -1:
            return text

        text = text[0:offset]+self.MORE_LINK.format(self.permalink, self.id)

        # return text
        return balance_tags(text)

    ## Add the read more anchor for the automatic mode.
    #
    # \param text The text of the entry
    # \return The text with the anchor, or the original text if [more] was not
    # found
    def applyFullAuto(self, text):
        offset = self._autoOffset(text)
        if offset != -1:
            return text[0:offset]+self.MORE_ANCHOR.format(self.id)+text[offset:]
        else:
            return text

    ## Get the offset at which the text should be cut in automatic mode.
    #
    # \param text The text to cut
    # \return See the description
    def _autoOffset(self, text):
        # We do not want to apply this mode if [more] was found in the entry.
        # In addition to that, we do not cut entries that would be too short.
        if text.find(self.MORE_TEXT) != -1 or self.manual or \
                len(text) < self.CHOP_MIN:
            return -1

        # Simpler algorithm than the one that I used on the original plugin:
        # just put the read more before tag from these one that is closest to
        # CHOP_AT.
        #
        # It is a bit hackish but it should work in most cases.
        find = ['</p>', '<br>', '<br />', '<br/>', '</div>', '</li>']
        min = len(text)
        for tag in find:
            tmp = text.find(tag, self.CHOP_AT)
            if tmp < min and tmp != -1:
                min = tmp

        return min
