## Some filters of WordPress that are used by FlatPress
#
# I tried to keep the code itself as the original one, but in my opinion some
# features were not used (as removal of <title>...</title> and
# <category>...</category>) or XHTML friendly filters (not needed with HTML 5).
#
# Finally note that some functions like wpautop are notoriously bugged and will
# not work for some cases (e.g. pre inside pre), however that was on FP code.
# In my entries they did not create problems, so I am keeping them, at least for
# old entries.
#
# This code is based on code by the WordPress project and therefore it is
# released under the GNU GPL License, version 2.

import re

_compiled_pre_re = re.compile('(<pre[^>]*>)(.*?)</pre>', re.I | re.S)

## Python implementation of wpautop
#
# \param pee The text which has to be formatted.
# \param br If set, this will convert all remaining line-breaks after
#           paragraphing. Default true.
# \link https://github.com/RealGeeks/wp_export_parser/blob/master/wp_export_parser/autop.py
def wpautop(pee, br=True):
    if not pee or pee.strip() == '':
            return ''
    pee = pee + "\n"  # just to make things a little easier, pad the end
    pee = re.sub('<br />\s*<br />', "\n\n", pee)

    # Space things ut a little
    allblocks = '(?:table|thead|tfoot|caption|col|colgroup|tbody|tr|td|th|div|dl|dd|dt|ul|ol|li|pre|select|form|map|area|blockquote|address|math|style|input|p|h[1-6]|hr|fieldset|legend)'
    pee = re.sub('(<' + allblocks + '[^>]*>)', r"\n\1", pee)
    pee = re.sub('(</' + allblocks + '>)', r"\1\n\n", pee)
    # Cross-platform newlines
    pee = pee.replace("\r\n", "\n")
    pee = pee.replace("\r", "\n")

    if pee.find('<object') != -1:
        pee = re.sub('\s*<param([^>]*)>\s*', r"<param\1>", pee)  # no pee inside object/embed
        pee = re.sub('\s*</embed>\s*', '</embed>', pee)
    pee = re.sub("/\n\n+/", "\n\n", pee)  # take care of duplicates

    # make paragraphs, including one at the end
    pees = re.split('\n\s*\n', pee)
    pee = ''
    for tinkle in pees:
        pee += '<p>' + tinkle.strip("\n") + "</p>\n"
    pee = re.sub('<p>\s*</p>', '', pee)  # under certain strange conditions it could create a P of entirely whitespace
    pee = re.sub('<p>([^<]+)</(div|address|form)>', r"<p>\1</p></\2>", pee)
    pee = re.sub('<p>\s*(</?' + allblocks + '[^>]*>)\s*</p>', r"\1", pee)  # don't pee all over a tag
    pee = re.sub("<p>(<li.+?)</p>", r"\1", pee)  # problem with nested lists
    pee = re.sub('<p><blockquote([^>]*)>', r"<blockquote\1><p>", pee)
    pee = pee.replace('</blockquote></p>', '</p></blockquote>')
    pee = re.sub('<p>\s*(</?' + allblocks + '[^>]*>)', r"\1", pee)
    pee = re.sub('(</?' + allblocks + '[^>]*>)\s*</p>', r"\1", pee)
    if br:
        pee = re.sub(r'<(script|style).*?</\1>', _autop_newline_preservation_helper, pee, flags=re.S)
        pee = re.sub('(?<!<br />)\s*\n', "<br />\n", pee)  # optionally make line breaks
        pee = pee.replace('<WPPreserveNewline />', "\n")
    pee = re.sub('(</?' + allblocks + '[^>]*>)\s*<br />', r"\1", pee)
    pee = re.sub('<br />(\s*</?(?:p|li|div|dl|dd|dt|th|pre|td|ul|ol)[^>]*>)', r'\1', pee)
    if pee.find('<pre') != -1:
        pee = _compiled_pre_re.sub(_clean_pre, pee)
    pee = re.sub("\n</p>$", '</p>', pee)

    return pee

## A helper that cleans pre for wpautop
#
# \param matches The matches of the regex that looks for pre
# \return The modified text
def _clean_pre(matches):
    text = matches.group(0)
    text = re.sub('<br />', '', text)
    text = re.sub('<br/>', '', text)
    text = re.sub('<br>', '', text)
    text = re.sub('<p>', '\n', text)
    text = re.sub('</p>', '', text)

    return text

## A helper that helps preserving style and script elements in wpautop
#
# \param matches The matches of the regex that looks for styles and scripts
# \return The modified text
def _autop_newline_preservation_helper(matches):
    return matches.group(0).replace("\n", "<WPPreserveNewline />")

_amp = re.compile('&([^#])(?![a-z]{1,8};)', re.I)

## Improve the typography of text
#
# Almost all variable names and comments of the function are the ones of the
# original PHP one.
# I ported almost everything as it was written in PHP, but I changed some lines
# with some solutions that were nicer, in my opinion
#
# \param text The text to improve
# \return text with improved typography
def wptexturize(text):
    output = ''

    # Capture tags and everything inside them
    textarr = re.split('(<.*?>)', text, 0, re.S)

    # Loop stuff
    skip = 0

    for curl in textarr:
        if curl and curl[0] != '<' and skip == 0:
            # If it's not a tag
            curl = curl.replace('---', '&#8212;');
            curl = curl.replace(' -- ', ' &#8212; ');
            curl = curl.replace('--', '&#8211;');
            curl = curl.replace('xn&#8211;', 'xn--');
            curl = curl.replace('...', '&#8230;');
            curl = curl.replace('``', '&#8220;');

            # This is a hack, look at this more later.
            # It works pretty well though.
            curl = curl.replace("'tain't", '&#8217;tain&#8217;t')
            curl = re.sub(r'\'(twere|twas|tis|twill|til|bout|nuff|round|cause)',
                    r'&#8217;\1', curl)

            curl = curl.replace('\'s', '&#8217;s')
            curl = re.sub(r'(\d\d(?:&#8217;|\')?s)', r'&#8217;\1', curl)
            curl = re.sub(r'(\s|^|")\'', r'\1&#8216;', curl)
            # curl = re.sub(r'(\d+)"', r'\1&#8243;', curl)

            # The PHP function included a $3, however this is a negative
            # lookahead, so Python complains if you include a \3 (and this
            # makes sense)
            curl = re.sub(r'(\s|^)("|&quot;)(?!\s)', r'\1&#8220;', curl)

            curl = re.sub('(\d+)\'', r'\1&#8242;', curl)
            curl = re.sub(r'(\S)\'([^\'\s])', r'\1&#8217;\2', curl)
            # curl = re.sub('/(\s|^)"(?!\s)/', '$1&#8220;$2', curl)

            # Note: this was repeated in the original PHP function and same
            # considerations apply as above
            curl = re.sub(r'(\s|^)("|&quot;)(?!\s)', r'\1&#8220;', curl)

            curl = re.sub(r'("|&quot;)(\s|\S|$)', r'&#8221;\2', curl)
            curl = re.sub(r'\'([\s.]|$)', r'&#8217;\1', curl)
            curl = re.sub(r'\(tm\)', ' &#8482;', curl, 0, re.I)
            curl = curl.replace("''", '&#8221;')

            curl = re.sub(r'(\d+)x(\d+)', r'\1&#215;\2', curl);

        elif curl.startswith('</') or curl.endswith('/>'):
            # Note (PieroV): in the original code strstr was used, which means
            # that the literal Python operations would be
            #       '</' in curl or '/>' in curl
            # but I really dislike this idea, and if the HTML is valid this
            # should work, whereas if the HTML is not valid, probably the
            # original function would fail, too.
            if skip > 0:
                skip -= 1

        elif curl.startswith(('<code', '<pre', '<kbd', '<style', '<script')):
            # As above, in the original code strstr was used, but with correct
            # HTML (tags not starting inside other tags), these strings can be
            # only at the beginning of the string.
            skip += 1
        elif curl and curl[0] == '<' and skip > 0:
            skip += 1

        curl = _amp.sub(r'&#038;\1', curl)
        output += curl

    return output


## Replaces Windows CP1252 specific characters and '&' that are alone
#
# \param content The string which the function have to work in
# \return The passed string modified
def convert_chars(content):
    # These are Windows CP1252 specific characters that would look weird on
    # non-Windows browsers.
    wp_htmltranswinuni = {
        '&#128;': '&#8364;', # the Euro sign
        '&#129;': '',
        '&#130;': '&#8218;',
        '&#131;': '&#402;',
        '&#132;': '&#8222;',
        '&#133;': '&#8230;',
        '&#134;': '&#8224;',
        '&#135;': '&#8225;',
        '&#136;': '&#710;',
        '&#137;': '&#8240;',
        '&#138;': '&#352;',
        '&#139;': '&#8249;',
        '&#140;': '&#338;',
        '&#141;': '',
        '&#142;': '&#382;',
        '&#143;': '',
        '&#144;': '',
        '&#145;': '&#8216;',
        '&#146;': '&#8217;',
        '&#147;': '&#8220;',
        '&#148;': '&#8221;',
        '&#149;': '&#8226;',
        '&#150;': '&#8211;',
        '&#151;': '&#8212;',
        '&#152;': '&#732;',
        '&#153;': '&#8482;',
        '&#154;': '&#353;',
        '&#155;': '&#8250;',
        '&#156;': '&#339;',
        '&#157;': '',
        '&#158;': '',
        '&#159;': '&#376;'
    }

    # Converts lone & characters into &#38; (a.k.a. &amp;)
    content = _amp.sub(r'&#038;\1', content)

    for (orig, replacement) in wp_htmltranswinuni.items():
        content = content.replace(orig, replacement)

    return content

## Known single-entity/self-closing tags
SINGLE_TAGS = ['area', 'base', 'basefont', 'br', 'col', 'command', 'embed',
               'frame', 'hr', 'img', 'input', 'isindex', 'link', 'meta',
               'param', 'source']
## Tags that can be immediately nested within themselves
NESTABLE_TAGS = ['blockquote', 'div', 'object', 'q', 'span']

## Balances tags of string using a modified stack.
#
# \since 2.0.4
#
# \author Leonard Lin <leonard@acm.org>
# \license GPL
# \copyright November 4, 2001
# \version 1.1
# \todo Make better - change loop condition to $text in 1.2
# \internal Modified by Scott Reilly (coffee2code) 02 Aug 2004
#		1.1  Fixed handling of append/stack pop order of end text
#			 Added Cleaning Hooks
#		1.0  First Version
#
# \param string $text Text to be balanced.
# \return string Balanced text.
def balance_tags(text):
    tagstack = []
    tagqueue = ''
    newtext = ''

    # WP bug fix for comments - in case you REALLY meant to type '< !--'
    text = text.replace('< !--', '<    !--')
    # WP bug fix for LOVE <3 (and other situations with '<' before a number)
    text = re.sub('<([0-9]{1})', r'&lt;\1', text)

    while True:
        m = re.search(r'<(\/?[\w:]*)\s*([^>]*)>', text)
        if not m:
            break
        newtext += tagqueue

        i = text.find(m.group(0))
        l = len(m.group(0))

        # clear the shifter
        tagqueue = ''
        # Pop or Push
        if m.group(1) and '/' == m.group(1)[0]: # End Tag
            tag = m.group(1)[1:].lower()
            # if too many closing tags
            if not tagstack:
                tag = ''
                # or close to be safe $tag = '/' . $tag;
            # if stacktop value = tag close value then pop
            elif tagstack[-1] == tag: # found closing tag
                tag = '</'+tag+'>' # Close Tag
                # Pop
                tagstack.pop()
            else: # closing tag not at top, search for it
                for i in range(len(tagstack)-1, -1, -1):
                    if tagstack[i] == tag:
                    # add tag to tagqueue
                        for k in range(len(tagstack)-1, j-1, -1):
                            tagqueue += '</'+tagstack.pop()+'>'
                        break

                tag = ''
        else: # Begin Tag
            tag = m.group(1).lower()
            attributes = m.group(2)

            # Tag Cleaning

            # If it's an empty tag "< >", do nothing
            if not tag:
                pass
            # ElseIf it presents itself as a self-closing tag...
            elif attributes and attributes[-1] == '/':
                # ...but it isn't a known single-entity self-closing tag, then
                # don't let it be treated as such and immediately close it with
                # a closing tag (the tag will encapsulate no text as a result)
                if tag not in SINGLE_TAGS:
                    attributes = attributes[0:-1].strip()+'></'+tag
            # ElseIf it's a known single-entity tag but it doesn't close
            # itself, do so
            elif tag in SINGLE_TAGS:
                attributes += '/'
            # Else it's not a single-entity tag
            else:
                # If the top of the stack is the same as the tag we want to
                # push, close previous tag
                if tagstack and tag not in NESTABLE_TAGS and \
                        tagstack[-1] == tag:
                    tagqueue = '</'+tagstack.pop()+'>'
                tagstack.append(tag)
            # Attributes

            if attributes and attributes[0] != '>':
                attributes = ' '+attributes

            tag = '<'+tag+attributes+'>'
            # If already queuing a close tag, then put this tag on, too
            if tagqueue:
                tagqueue += tag
                tag = ''

        newtext += text[0:i]+tag
        text = text[i+l:]

    # Clear Tag Queue
    newtext += tagqueue

    # Add Remaining text
    newtext += text

    # Empty Stack
    while tagstack:
        x = tagstack.pop()
        newtext += '</'+x+'>' # Add remaining tags to close

    # WP fix for the bug with HTML comments
    newtext = newtext.replace('< !--', '<!--')
    newtext = newtext.replace('<    !--', '< !--')

    return newtext
